//=============================================================================
//
// FSSound module for MS Flight Simulator 98, 2000 and Combat Flight Simulator
// Copyright (c) 1999-2000 by Daniel Steiner <dstbon@bigfoot.com>
//
// This module allows you to use sound files in FS gauges trough an easy to use
// interface.
//
// Note: to compile, the following macros must be defined on the comand line:
//    # DirectSound version 5 is used for compatibility with FS98 and CFS.
//    DIRECTSOUND_VERSION=0x0500
//
//    # Name of the window class of the Flight Simulator main window.
//    # This is 'FS98MAIN' for FS98, CFS and FS2000.
//    FSWND="FS98MAIN"
//
//    # Name of the sound resources.
//    # The sound modules finds only resources of this type.
//    SOUND_RES="SOUND"
//
//    # Optional: to get some debug output
//    #DEBUG
//
//=============================================================================
#include "FSSound.h"

#define DLLEXPORT __declspec(dllexport)
#define FSAPI __stdcall


//-----------------------------------------------------------------------------
// set up the interface to Flight Simulator
// simply let FS ingnore this module, but if we don't provide these exports,
// FS will crash during initlialization
//-----------------------------------------------------------------------------
DLLEXPORT struct {
	struct {
		int fnID;
		PVOID fnptr;
	} IMPORTSentry;
	struct {
		int fnID;
		PVOID fnptr;
	} dummyentry;
} ImportTable = {
	{0x00000000, NULL},
	{0x00000000, NULL}
};

void FSAPI module_init(void) {}
void FSAPI module_deinit(void) {}

DLLEXPORT struct {
	int ModuleID;
	void (FSAPI *ModuleInit)(void);
	void (FSAPI *ModuleDeinit)(void);
	UINT ModuleFlags;
	UINT ModulePriority;
	PVOID DialogTable;
	PVOID gauge_header_ptr;
} Linkage = {
	0x00000000,
	module_init,
	module_deinit,
	0,
	0,
	NULL,
	NULL
};


//-----------------------------------------------------------------------------
// golbal vars
//-----------------------------------------------------------------------------
LPDIRECTSOUND lpds = NULL;	// holds the DirectSound interface
LPFSSOUND list = NULL;		// holds a list of all registered sound buffers
BOOL init = FALSE;			// have we initialized DirectSound?
HWND hwnd = NULL;			// holds the window handle of the parent app


//-----------------------------------------------------------------------------
// the prototypes for internal procedures
//-----------------------------------------------------------------------------
void ReleaseDSound(void);	// free the list of all sound buffers and release
							// uninitialize DirectSound
void InitDSound(void);		// initialize DirectSound and set up the golbal vars
DWORD CodeString(PCHAR);	// calculates an integer from a string
DWORD CodeIntegers(DWORD,
	DWORD);					// calculates an integer from two integers
BOOL LoadWaveFile(HMMIO,
	LPLPDIRECTSOUNDBUFFER);	// load the wave file/res from the HMMIO handle to
							// the DirectSound buffer
BOOL DuplicateBuffer(LPLPFSSOUND,
	DWORD);					// try to duplicate an existing sound buffer

//-----------------------------------------------------------------------------
// let the preprocessor accept strings from the command line
//-----------------------------------------------------------------------------
#define xquote(X) #X
#define quote(X) xquote(X)


//-----------------------------------------------------------------------------
// some debugging stuff
//-----------------------------------------------------------------------------
#ifdef DEBUG
#define log(TEXT_OUT)	OutputDebugString(TEXT_OUT"\n")
#else
#define log(X)
#endif


//-----------------------------------------------------------------------------
// exported procedures
//-----------------------------------------------------------------------------
DLLEXPORT BOOL RegisterFile(LPLPFSSOUND lplpFSSound, PCHAR lpszName) {
	HMMIO hmmio = NULL;
	LPDIRECTSOUNDBUFFER lpdsbStatic;
	DWORD dwCode;

	log("RegisterFile()");
	// check wheter we need to initialize DirectSound
	if (!init)
		InitDSound();

	// check valid parameters and sound available
	if (!lpds || !lplpFSSound || !lpszName) {
		log("invalid lpds or lplpFSSound or lpszName");
		return FALSE;
	}

	// unregister old sound
	if (*lplpFSSound) {
		log("first unregistering lplpFSSound");
		Unregister(lplpFSSound);
	}
	else {
		*lplpFSSound = NULL;
	}

	// check whether we can just duplicate an existsting buffer
	dwCode = CodeString(lpszName);
	if (DuplicateBuffer(lplpFSSound, dwCode)) {
		log("RegisterFile() completed, with buffer duplicating");
		return TRUE;
	}

	// let's do it the hard way
	log("no buffer found to duplicate, creating a new one");
	// open the wave file
	if (!(hmmio = mmioOpen(lpszName, NULL, MMIO_ALLOCBUF | MMIO_READ))) {
		log("mmioOpen() failed");
		return FALSE;
	}

	// read in the wave file
	log("Reading in the wave file");
	if (LoadWaveFile(hmmio, &lpdsbStatic) == FALSE) {
		log("Reading in the wave file failed");
		return FALSE;
	}
	log("Reading in the wave file completed");

	// create the return structure
	if (!(*lplpFSSound = (LPFSSOUND)malloc(sizeof(FSSOUND)))) {
		IDirectSoundBuffer_Release(lpdsbStatic);
		log("malloc() for list element failed\nreleasing DirectSoundBuffer Interface");
		return FALSE;
	}

	// fill in the data
	(*lplpFSSound)->next = list;
	(*lplpFSSound)->lpDsb = lpdsbStatic;
	(*lplpFSSound)->code = dwCode;
	list = *lplpFSSound;
	log("RegisterFile() completed, with new buffer");
	return TRUE;

}

DLLEXPORT BOOL RegisterRes(LPLPFSSOUND lplpFSSound, UINT res, HMODULE module) {
	HMMIO hmmio = NULL;
	LPDIRECTSOUNDBUFFER lpdsbStatic;
	HRSRC hrsrc;
	HGLOBAL hglobal;
	LPVOID resdata;
	MMIOINFO mmioinfo;
	DWORD dwCode;

	log("RegisterRes()");
	// check wheter we need to initialize DirectSound
	if (!init)
		InitDSound();

	// check valid parameters and sound available
	if (!lpds || !lplpFSSound || !module) {
		log("invalid lpds or lplpFSSound or lpszName");
		return FALSE;
	}

	// unregister old sound
	if (*lplpFSSound) {
		log("first unregistering lplpFSSound");
		Unregister(lplpFSSound);
	}
	else {
		*lplpFSSound = NULL;
	}

	// check whether we can just duplicate an existsting buffer
	dwCode = CodeIntegers(res, module);
	if (DuplicateBuffer(lplpFSSound, dwCode)) {
		log("RegisterRes() completed, with buffer duplicating");
		return TRUE;
	}

	// let's do it the hard way
	log("no buffer found to duplicate, creating a new one");
	// find the resource
	if (!(hrsrc = FindResource(module, res, quote(SOUND_RES)))) {
		log("FindResource() failed");
		return FALSE;
	}

	// load the resource into memory
	if (!(hglobal = LoadResource(module, hrsrc))) {
		log("LoadResource() failed");
		return FALSE;
	}

	// lock the resource data in memory
	if (!(resdata = LockResource(hglobal))) {
		FreeResource(hglobal);
		log("LockResource() failed\nfreeing resource");
		return FALSE;
	}

	// initialize the MMIOINFO structure
	memset(&mmioinfo, 0, sizeof(MMIOINFO));
	mmioinfo.pchBuffer = resdata;
	mmioinfo.fccIOProc = FOURCC_MEM;
	mmioinfo.cchBuffer = SizeofResource(module, hrsrc);

	// get a multi media file handle to the sound data
	if (!(hmmio = mmioOpen(NULL, &mmioinfo, MMIO_READ))) {
		UnlockResource(resdata);
		FreeResource(hglobal);
		log("mmioOpen()failed\nunlocking and freeing resource");
		return FALSE;
	}

	// read in the wave file
	if (!LoadWaveFile(hmmio, &lpdsbStatic)) {
		UnlockResource(resdata);
		FreeResource(hglobal);
		log("Reading in the wave file failed\nunlocking and freeing resource");
		return FALSE;
	}

	// unlock and free the resource
	log("unlocking and freeing resource");
	UnlockResource(resdata);
	FreeResource(hglobal);

	// create the return structure
	if (!(*lplpFSSound = (LPFSSOUND)malloc(sizeof(FSSOUND)))) {
		IDirectSoundBuffer_Release(lpdsbStatic);
		log("malloc() for list element failed\nreleasing DirectSoundBuffer Interface");
		return FALSE;
	}

	// fill in the data
	(*lplpFSSound)->next = list;
	(*lplpFSSound)->lpDsb = lpdsbStatic;
	(*lplpFSSound)->code = dwCode;
	list = *lplpFSSound;
	log("RegisterRes() completed, with new buffer");
	return TRUE;
}


DLLEXPORT void Unregister(LPLPFSSOUND lplpFSSound) {
	LPFSSOUND cur;

	log("Unregister()");

	// check valid
	if (lplpFSSound && *lplpFSSound) {
		log("scanning list");

		// remove the element from the list
		if (list == *lplpFSSound) {
			list = (*lplpFSSound)->next;  // handle first element
			log("first element found and unchained");
		}
		else
			for(cur = list; cur; cur = cur->next) {
				if (cur->next == *lplpFSSound) {
					cur->next = (*lplpFSSound)->next; // scan the rest of the list
					log("element found and unchained");
					break;
				}
			}

		// release the DirectSoundBuffer
		if ((*lplpFSSound)->lpDsb) {
			log("releasing DirectSoundBuffer Interface");
			IDirectSoundBuffer_Release((*lplpFSSound)->lpDsb);
		}

		// free the memory
		log("freeing the memory");
		free((*lplpFSSound));
		*lplpFSSound = NULL;
	}

	// uninitialize DirectSound if there are no more sound buffers
	if (!list)
		ReleaseDSound();
	log("Unregister() completed");
}

DLLEXPORT BOOL Play(LPFSSOUND lpFSSound, BOOL loop) {
	DWORD flags = 0;

	log("Play()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if (loop) {
		log("setting DSBPLAY_LOOPING flag");
		flags = DSBPLAY_LOOPING;
	}
	if (IDirectSoundBuffer_Play(lpFSSound->lpDsb, 0, 0, flags) == DS_OK) {
		log("Play() completed");
		return TRUE;
	}
	log("Play() failed");
	return FALSE;
}

DLLEXPORT BOOL Stop(LPFSSOUND lpFSSound) {
	log("Stop()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if ((IDirectSoundBuffer_Stop(lpFSSound->lpDsb) == DS_OK)
		&& (IDirectSoundBuffer_SetCurrentPosition(lpFSSound->lpDsb, 0) == DS_OK)) {
		log("Stop() completed");
		return TRUE;
	}
	log("Stop() or SetCurrentPosition() failed");
	return FALSE;
}

DLLEXPORT BOOL IsPlaying(LPFSSOUND lpFSSound) {
	DWORD status;

	log("IsPlaying()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if ((IDirectSoundBuffer_GetStatus(lpFSSound->lpDsb, &status) == DS_OK)
		&& ((status & DSBSTATUS_PLAYING))) {
		log("IsPlaying() completed");
		return TRUE;
	}
	log("GetStatus() failed or buffer not playing");
	return FALSE;
}

DLLEXPORT LONG GetVolume(LPFSSOUND lpFSSound) {
	LONG volume;

	log("GetVolume()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return DSBVOLUME_MIN;
	}
	if (IDirectSoundBuffer_GetVolume(lpFSSound->lpDsb, &volume) == DS_OK) {
		log("GetVolume() completed");
		return volume;
	}
	log("GetVolume() failed");
	return DSBVOLUME_MIN;
}

DLLEXPORT BOOL SetVolume(LPFSSOUND lpFSSound, LONG volume) {

	log("SetVolume()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if (IDirectSoundBuffer_SetVolume(lpFSSound->lpDsb, volume) == DS_OK) {
		log("SetVolume() completed");
		return TRUE;
	}
	if ((volume > DSBVOLUME_MAX) || (volume < DSBVOLUME_MIN)) {
		log("paramter out of range");
	}
	log("SetVolume() failed");
	return FALSE;
}

DLLEXPORT LONG GetPan(LPFSSOUND lpFSSound) {
	LONG pan;

	log("GetPan()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return DSBPAN_CENTER;
	}
	if (IDirectSoundBuffer_GetPan(lpFSSound->lpDsb, &pan) == DS_OK) {
		log("GetPan() completed");
		return pan;
	}
	log("GetPan() failed");
	return DSBPAN_CENTER;
}

DLLEXPORT BOOL SetPan(LPFSSOUND lpFSSound, LONG pan) {

	log("SetPan()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if (IDirectSoundBuffer_SetPan(lpFSSound->lpDsb, pan) == DS_OK) {
		log("SetPan() completed");
		return TRUE;
	}
	if ((pan > DSBPAN_RIGHT) || (pan < DSBPAN_LEFT)) {
		log("paramter out of range");
	}
	log("SetPan() failed");
	return FALSE;
}

DLLEXPORT DWORD GetFrequency(LPFSSOUND lpFSSound) {
	DWORD frequency;

	log("GetFrequency()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return DSBFREQUENCY_ORIGINAL;
	}
	if (IDirectSoundBuffer_GetFrequency(lpFSSound->lpDsb, &frequency) == DS_OK) {
		log("GetFrequency() completed");
		return frequency;
	}
	log("GetFrequency() failed");
	return DSBFREQUENCY_ORIGINAL;
}

DLLEXPORT BOOL SetFrequency(LPFSSOUND lpFSSound, DWORD frequency) {

	log("SetFrequency()");

	if (!lpFSSound || !lpFSSound->lpDsb) {
		log("invalid lpFSSound or lpFSSound->lpDsb");
		return FALSE;
	}
	if (IDirectSoundBuffer_SetFrequency(lpFSSound->lpDsb, frequency) == DS_OK) {
		log("SetFrequency() completed");
		return TRUE;
	}
	if ((frequency > DSBFREQUENCY_MAX) || (frequency < DSBFREQUENCY_MIN)) {
		log("paramter out of range");
	}
	log("SetFrequency() failed");
	return FALSE;
}

//-----------------------------------------------------------------------------
// local procedures
//-----------------------------------------------------------------------------
BOOL DuplicateBuffer(LPLPFSSOUND lplpFSSound, DWORD dwCode) {
	LPFSSOUND cur;
	DWORD cT;
	LPDIRECTSOUNDBUFFER lpdsbStatic;

	// scan the list of sound buffers
	log("searching for buffer to duplicate");
	for (cur = list; cur; cur = cur->next) {
		// have we found a coresponding structure?
		if ((cur->code == dwCode)
			&& (cur->lpDsb)
			&& (IDirectSoundBuffer_GetStatus(cur->lpDsb, &cT) == DS_OK)
			&& (!(cT & DSBSTATUS_BUFFERLOST))) {
			// create the sound buffer
			log("found valid buffer to duplicate");
			if (IDirectSound_DuplicateSoundBuffer(lpds, cur->lpDsb,
				&lpdsbStatic) != DS_OK) {
				log("DuplicateSoundBuffer() failed");
				return FALSE;
			}

			// create the return structure
			if (!(*lplpFSSound = (LPFSSOUND)malloc(sizeof(FSSOUND)))) {
				log("malloc() failed, releasing DirectSoundBuffer Interace");
				IDirectSoundBuffer_Release(lpdsbStatic);
				return FALSE;
			}

			// fill in the data
			(*lplpFSSound)->next = list;
			(*lplpFSSound)->lpDsb = lpdsbStatic;
			(*lplpFSSound)->code = dwCode;
			list = *lplpFSSound;
			return TRUE;
		}
	}
	log("no valid buffer found to duplicate");
	return FALSE;
}

BOOL LoadWaveFile(HMMIO hmmio, LPLPDIRECTSOUNDBUFFER lplpdsbStatic) {
	MMCKINFO ckInfo, ckInfoParent;
	MMIOINFO mmioinfo;
	PCMWAVEFORMAT pcmWaveFormat;
	WAVEFORMATEX *pwfx = NULL;
	DSBUFFERDESC dsbdesc;
	LPVOID lpvAudio1;
	DWORD  dwBytes1, cT;

	// read the parent structure
	if ((mmioDescend(hmmio, &ckInfoParent, NULL, 0) != 0)
		|| (ckInfoParent.ckid != FOURCC_RIFF)
		|| (ckInfoParent.fccType != mmioFOURCC('W', 'A', 'V', 'E'))) {
		mmioClose(hmmio, 0);
		log("mmioDescend() failed or invalid file type\nclosing file");
		return FALSE;
	}


	// get the wave format data
	ckInfo.ckid = mmioFOURCC('f', 'm', 't', ' ');
	if ((mmioDescend(hmmio, &ckInfo, &ckInfoParent, MMIO_FINDCHUNK) != 0)
		|| (ckInfo.cksize < (LONG)sizeof(PCMWAVEFORMAT))
		|| (mmioRead(hmmio, (HPSTR) &pcmWaveFormat, sizeof(pcmWaveFormat))
			!= sizeof(pcmWaveFormat))) {
		mmioClose(hmmio, 0);
		log("mmioDescend() failed or invalid wave format\nclosing file");
		return FALSE;
	}

	// fill the WAVEFORMEX structure
	if(pcmWaveFormat.wf.wFormatTag == WAVE_FORMAT_PCM) {
		if(!(pwfx = (WAVEFORMATEX *)malloc(sizeof(WAVEFORMATEX)))) {
			mmioClose(hmmio, 0);
			log("malloc() for WAVEFORMATEX failed\nclosing file");
			return FALSE;
		}

		// copy the structure
		memcpy(pwfx, &pcmWaveFormat, sizeof(pcmWaveFormat));
		pwfx->cbSize = 0;
	}
	else {
		// Read in length of extra bytes.
		WORD cbExtraBytes = 0L;
		if ((mmioRead(hmmio, (CHAR*)&cbExtraBytes, sizeof(WORD)) != sizeof(WORD))
			|| (!(pwfx = (WAVEFORMATEX *)malloc(sizeof(WAVEFORMATEX)
				+ cbExtraBytes)))) {
			mmioClose(hmmio, 0);
			log("mmioRead() or malloc() for WAVEFORMATEX failed\nclosing file");
			return FALSE;
		}

		// copy the structure
		memcpy(pwfx, &pcmWaveFormat, sizeof(pcmWaveFormat));
		pwfx->cbSize = cbExtraBytes;

		// Now, read those extra bytes into the structure
		if(mmioRead(hmmio, (CHAR*)(((BYTE*)&(pwfx->cbSize)) + sizeof(WORD)),
			cbExtraBytes) != cbExtraBytes) {
			free(pwfx);
			pwfx = NULL;
			mmioClose(hmmio, 0);
			log("mmioRead() failed\nclosing file, freeing memory");
			return FALSE;
		}
	}

	// get out of the format part and find the data part
	if ((mmioAscend(hmmio, &ckInfo, 0) != 0)
		|| (mmioSeek(hmmio, ckInfoParent.dwDataOffset + sizeof(FOURCC),
			SEEK_SET) == -1)) {
		free(pwfx);
		pwfx = NULL;
		mmioClose(hmmio, 0);
		log("mmioAscend() or mmioSeek() failed\nclosing file, freeing memory");
		return FALSE;
	}

	// search the data part
	ckInfo.ckid = mmioFOURCC('d', 'a', 't', 'a');
	if(mmioDescend(hmmio, &ckInfo, &ckInfoParent, MMIO_FINDCHUNK) != 0)  {
		free(pwfx);
		pwfx = NULL;
		mmioClose(hmmio, 0);
		log("mmioDescend() failed\nclosing file, freeing memory");
		return FALSE;
	}

	// initialize the DSBUFFERDESC structure
	memset(&dsbdesc, 0, sizeof(DSBUFFERDESC));
	dsbdesc.dwSize = sizeof(DSBUFFERDESC);
	dsbdesc.dwFlags = DSBCAPS_STATIC | DSBCAPS_CTRLVOLUME;
	#ifdef DEBUG
		dsbdesc.dwFlags |= DSBCAPS_GLOBALFOCUS;
	#endif
	dsbdesc.dwBufferBytes = ckInfo.cksize;
	dsbdesc.lpwfxFormat = pwfx;

	// create the sound buffer
	if (IDirectSound_CreateSoundBuffer(lpds, &dsbdesc, lplpdsbStatic, NULL)) {
		free(pwfx);
		pwfx = NULL;
		mmioClose(hmmio, 0);
		log("CreateSoundBuffer() failed\nclosing file, freeing memory");
		return FALSE;
	}

	// free the WAVEFOTMATEX structure that is no longer needed
	free(pwfx);
	pwfx = NULL;

	// get a write lock on the sound buffer
	if (IDirectSoundBuffer_Lock(*lplpdsbStatic, 0, 0, &lpvAudio1, &dwBytes1,
		NULL, NULL, DSBLOCK_ENTIREBUFFER) != DS_OK) {
		mmioClose(hmmio, 0);
		IDirectSoundBuffer_Release(*lplpdsbStatic);
		log("Lock() failed\nclosing file, releasing DirectSoundBuffer Interface");
		return FALSE;
	}

	if(dwBytes1 > ckInfo.cksize)
		dwBytes1 = ckInfo.cksize;
	ckInfo.cksize -= dwBytes1;

	// initialize the MMINFO structure
	if(mmioGetInfo(hmmio, &mmioinfo, 0) != 0) {
		mmioClose(hmmio, 0);
		IDirectSoundBuffer_Release(*lplpdsbStatic);
		log("mmioGetInfo() failed\nclosing file, releasing DirectSoundBuffer Interface");
		return FALSE;
	}

	// read the sound data into the buffer
	for(cT = 0; cT < dwBytes1; cT++) {
		if (mmioinfo.pchNext == mmioinfo.pchEndRead) {
			if ((mmioAdvance(hmmio, &mmioinfo, MMIO_READ) != 0)
				|| (mmioinfo.pchNext == mmioinfo.pchEndRead)) {
				mmioClose(hmmio, 0);
				IDirectSoundBuffer_Release(*lplpdsbStatic);
				log("mmioAdvance() failed\nclosing file, releasing DirectSoundBuffer Interface");
				return FALSE;
			}
		}

		// copy the bytes
		*((BYTE*)lpvAudio1 + cT) = *((BYTE*)mmioinfo.pchNext);
		mmioinfo.pchNext++;
	}

	// close the wave file
	mmioClose(hmmio, 0);

	// unlock the sound buffer
	if (IDirectSoundBuffer_Unlock(*lplpdsbStatic, lpvAudio1, cT, NULL, 0)
		!= DS_OK) {
		IDirectSoundBuffer_Release(*lplpdsbStatic);
		log("Unlock() failed\nreleasing DirectSoundBuffer Interface");
		return FALSE;
	}

	// all worked well
	return TRUE;
}

DWORD CodeString(PCHAR lpszName) {
	DWORD c = 0, pos = 0;
	log("CodeString()");
	while (lpszName[pos] != '\0')
		c = 31 * c + lpszName[pos++];
	log("CodeString() completed");
	return c;
}

DWORD CodeIntegers(DWORD a, DWORD b) {
	log("CodeIntegers()");
	return a ^ b;
}

void InitDSound() {

	log("InitDSound()");

	if (DirectSoundCreate(NULL, &lpds, NULL) != DS_OK) {
		log("DirectSoundCreate() failed");
		lpds = NULL;
	}
	else if (IDirectSound_SetCooperativeLevel(lpds, hwnd, DSSCL_NORMAL) != DS_OK) {
		log("SetCooperativeLevel() failed\nreleasing DirectSound Interface");
		IDirectSound_Release(lpds);
		lpds = NULL;
	}
	if (lpds)
		init = TRUE;

	log("InitDSound() completed");
}

void ReleaseDSound() {
	LPFSSOUND cur, prev;

	log("ReleaseDSound()");

	// iterate through the list of sound descriptors and clean up
	for(cur = list, prev = NULL; cur; prev = cur, cur = cur->next) {
		// check and release the DirectSoundBuffer
		if (cur->lpDsb)
			IDirectSoundBuffer_Release(cur->lpDsb);
		// free the FSSOUND structure
		if (prev)
			free(prev);
		log("releasing DirectSoundBuffer Interface and freeing list element");
	}

	// free the last element
	if (prev)
		free(prev);

	// release the DirectSound Interface
	if (lpds) {
		log("releasing the DirectSound Interface");
		IDirectSound_Release(lpds);
		lpds = NULL;
	}
	log("ReleaseDSound() completed");
	init = FALSE;
}

//-----------------------------------------------------------------------------
// the dll main procedure
//-----------------------------------------------------------------------------
BOOL APIENTRY DllMain (HINSTANCE hInst, DWORD reason, LPVOID reserved) {
	switch (reason) {
		case DLL_PROCESS_ATTACH:
			DisableThreadLibraryCalls(hInst);
			hwnd = FindWindow(quote(FSWND), NULL);
			break;

		case DLL_PROCESS_DETACH:
			if (init)
				ReleaseDSound();
			break;

		case DLL_THREAD_ATTACH:
			break;

		case DLL_THREAD_DETACH:
			break;
	}
	return TRUE;
}
