//SPEED.CONV.c
//Copyright (c) 2000 Arne Bartels, Oldenburg,Germany
//Speed and conversion routines for adapted MS Flight Simulator gauges after :
//
//  P_Assist V1.5.xls  Brian Tooby 2000.
//

/*
Tokens and their conversion factors:
MACH	:	MODULE_VAR	mach={MACH};						mach.var_value.n*3.2/65536
TAT		:	MODULE_VAR	total={TOTAL_AIR_TEMP};				total.var_value.n*256
OAT		:	MODULE_VAR	ambient={AMBIENT_TEMP_DEGREES_C};	ambient.var_value.n*1
IAS		:	MODULE_VAR	airspeed={AIRSPEED};				airspeed.var_value.n*1
ALT		:	MODULE_VAR	alt={PLANE_ALTITUDE};				alt.var_value.n*3.28084
*/

FLOAT64 IsaSat(FLOAT64 dblAlt);
FLOAT64	TatFromOat(FLOAT64 dblOat,FLOAT64 dblMach);
FLOAT64	OatFromTat(FLOAT64 dblTat,FLOAT64 dblMach);
FLOAT64	PCorrFromAlt(FLOAT64 dblAlt);
FLOAT64	MachFromIas(FLOAT64 dblAlt,FLOAT64 dblIas);
FLOAT64	IasFromMach(FLOAT64 dblAlt,FLOAT64 dblMach);
FLOAT64	TasFromMachTat(FLOAT64 dblTatC,FLOAT64 dblMach);
FLOAT64	MachFromTasTat(FLOAT64 dblTatC,FLOAT64 dblTas);
FLOAT64	TasFromMachOat(FLOAT64 dblOatC,FLOAT64 dblMach);
FLOAT64	MachFromTasOat(FLOAT64 dblOatC,FLOAT64 dblTas);

FLOAT64 IsaSat(FLOAT64 dblAlt)
{
	//returns static air temperature (SAT) according to international standard atmaosphere (ISA)
	//input:
	//      altitude in feet
	if(dblAlt>36000)
		return -56.5;
	return 15.0-dblAlt/1000.0*1.98;

}

FLOAT64	TatFromOat(FLOAT64 dblOat,FLOAT64 dblMach)
{
	//returns total air temperature (TAT)
	//input: 
	//       outside air temperature (OAT) in degrees centigrade
	//       mach number
	//TAT is higher than OAT due to friction at aircraft body
	return (1+0.2*dblMach*dblMach)*(dblOat+273.15)-273.15;
}

FLOAT64	OatFromTat(FLOAT64 dblTat,FLOAT64 dblMach)
{
	//returns outside air temperature (OAT)
	//input: 
	//       total air temperature (TAT) in degrees centigrade
	//       mach number
	//OAT is lower than TAT due to friction at aircraft body
	return (dblTat+273.15)/(1+0.2*dblMach*dblMach)-273.15;
}

FLOAT64	PCorrFromAlt(FLOAT64 dblAlt)
{
	//returns pressure correction factor for following calculations
	//input: 
	//      altitude in feet
	if(dblAlt>140000)
		return 0;
	return pow(1.0-3.566*dblAlt/518670,-5.2563);
}

FLOAT64	MachFromIas(FLOAT64 dblAlt,FLOAT64 dblIas)
{
	//returns mach number 
	//input: 
	//       altitude in feet
	//       indicated airspeed in knots
	FLOAT64 PressureCorrection=PCorrFromAlt(dblAlt);
	FLOAT64 dblValue=0;
	dblValue=1+0.2*dblIas*dblIas/437582.25;
	if(dblValue<0)
		return 0;
	dblValue=sqrt(dblValue*dblValue*dblValue*dblValue*dblValue*dblValue*dblValue)-1;
	dblValue*=PressureCorrection;
	dblValue+=1.0;
	if(dblValue<0)
		return 0;
	dblValue=5*(pow(dblValue,2.0/7.0)-1.0);
	if(dblValue<0)
		return 0;
	return sqrt(dblValue);
}

FLOAT64	IasFromMach(FLOAT64 dblAlt,FLOAT64 dblMach)
{
	//returns indicated airspeed in knots
	//input: 
	//       altitude in feet
	//       mach number 
	FLOAT64 PressureCorrection=PCorrFromAlt(dblAlt);
	FLOAT64 dblValue=0;
	dblValue=1.0+0.2*dblMach*dblMach;
	dblValue=sqrt(dblValue*dblValue*dblValue*dblValue*dblValue*dblValue*dblValue)-1;
	dblValue/=PressureCorrection;
	dblValue+=1;
	if(dblValue<0)
		return 0;
	dblValue=5*(pow(dblValue,2.0/7.0)-1);
	if(dblValue<0)
		return 0;
	return sqrt(dblValue)*661.5;
}

FLOAT64	TasFromMachTat(FLOAT64 dblTatC,FLOAT64 dblMach)
{
	//returns true airspeed in knots
	//input:
	//      true air temperature in degrees centigrade
	//      mach number
	return 39.0*dblMach*sqrt((dblTatC+273.15)/(1+0.2*dblMach*dblMach));
}

FLOAT64	MachFromTasTat(FLOAT64 dblTatC,FLOAT64 dblTas)
{
	//returns mach number
	//input:
	//      true air temperature in degrees centigrade
	//      true airspeed in knots
	FLOAT64 dblValue=39.0*39.0/dblTas/dblTas*(dblTatC+273.15)-0.2;
	if(dblValue<=0)
		return 0;
	return 1.0/sqrt(dblValue);
}
FLOAT64	TasFromMachOat(FLOAT64 dblOatC,FLOAT64 dblMach)
{
	//returns true airspeed in knots
	//input:
	//      outside air temperature in degrees centigrade
	//      mach number
	return 39.0*dblMach*sqrt(dblOat+273.15);
}

FLOAT64	MachFromTasOat(FLOAT64 dblOatC,FLOAT64 dblTas)
{
	//returns mach number
	//input:
	//      outside air temperature in degrees centigrade
	//      true airspeed in knots
	FLOAT64 dblValue=dblOatC+273.15;
	if(dblValue<=0)
		return 0;
	return dblTas/39/sqrt(dblValue);
}
